"""REST API view tests for the filesfolders app"""

import json

from django.urls import reverse

from test_plus.test import APITestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_views_api import SODARAPIViewTestMixin
from projectroles.views_api import INVALID_PROJECT_TYPE_MSG

from filesfolders.tests.test_views import (
    ZIP_PATH_NO_FILES,
    FilesfoldersViewTestMixin,
)
from filesfolders.models import Folder, File, HyperLink
from filesfolders.views_api import (
    FILESFOLDERS_API_MEDIA_TYPE,
    FILESFOLDERS_API_DEFAULT_VERSION,
)


# SODAR constants
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']

# Local constants
INVALID_UUID = '11111111-1111-1111-1111-111111111111'


class FilesfoldersAPIViewTestBase(
    FilesfoldersViewTestMixin, SODARAPIViewTestMixin, APITestCase
):
    """Base class for filesfolders API tests"""

    media_type = FILESFOLDERS_API_MEDIA_TYPE
    api_version = FILESFOLDERS_API_DEFAULT_VERSION

    def setUp(self):
        super().setUp()
        # Get knox token for self.user
        self.knox_token = self.get_token(self.user)


class TestFolderListCreateAPIView(FilesfoldersAPIViewTestBase):
    """Tests for FolderListCreateAPIView"""

    def setUp(self):
        super().setUp()
        self.folder_data = {
            'name': 'New Folder',
            'flag': 'IMPORTANT',
            'description': 'Folder\'s description',
        }
        self.url = reverse(
            'filesfolders:api_folder_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_get_list(self):
        """Test FolderListCreateAPIView GET to list folders"""
        response = self.request_knox(self.url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'name': self.folder.name,
            'folder': None,
            'owner': str(self.folder.owner.sodar_uuid),
            'project': str(self.folder.project.sodar_uuid),
            'flag': self.folder.flag,
            'description': self.folder.description,
            'date_modified': self.get_drf_datetime(self.folder.date_modified),
            'sodar_uuid': str(self.folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), [expected])

    def test_get_pagination(self):
        """Test GET with pagination"""
        url = self.url + '?page=1'
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'count': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'name': self.folder.name,
                    'folder': None,
                    'owner': str(self.folder.owner.sodar_uuid),
                    'project': str(self.folder.project.sodar_uuid),
                    'flag': self.folder.flag,
                    'description': self.folder.description,
                    'date_modified': self.get_drf_datetime(
                        self.folder.date_modified
                    ),
                    'sodar_uuid': str(self.folder.sodar_uuid),
                }
            ],
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_post_create_root(self):
        """Test POST to create folder in root"""
        response = self.request_knox(
            self.url, method='POST', data=self.folder_data
        )
        self.assertEqual(response.status_code, 201, msg=response.data)
        new_folder = Folder.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_folder)
        expected = {
            **self.folder_data,
            'folder': None,
            'owner': str(self.user.sodar_uuid),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_folder.date_modified),
            'sodar_uuid': str(new_folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_post_create_folder(self):
        """Test POST to create folder under folder"""
        folder_data = {
            **self.folder_data,
            'folder': str(self.folder.sodar_uuid),
        }
        response = self.request_knox(self.url, method='POST', data=folder_data)
        self.assertEqual(response.status_code, 201, msg=response.data)
        new_folder = Folder.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_folder)
        expected = {
            **folder_data,
            'owner': str(self.user.sodar_uuid),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_folder.date_modified),
            'sodar_uuid': str(new_folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_post_create_category(self):
        """Test POST to create folder in category (should fail)"""
        category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.make_assignment(category, self.user, self.role_owner)
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_list_create',
                kwargs={'project': category.sodar_uuid},
            ),
            method='POST',
            data=self.folder_data,
        )
        self.assertEqual(response.status_code, 403, msg=response.data)
        self.assertEqual(
            str(response.data['detail']),
            INVALID_PROJECT_TYPE_MSG.format(project_type=PROJECT_TYPE_CATEGORY),
        )


class TestFolderRetrieveUpdateDestroyAPIView(FilesfoldersAPIViewTestBase):
    """Tests for FolderRetrieveUpdateDestroyAPIView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': self.folder.sodar_uuid},
        )

    def test_get_retrieve(self):
        """Test FolderRetrieveUpdateDestroyAPIView GET to retrieve folder"""
        response = self.request_knox(self.url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'name': self.folder.name,
            'folder': None,
            'owner': str(self.folder.owner.sodar_uuid),
            'project': str(self.folder.project.sodar_uuid),
            'flag': self.folder.flag,
            'description': self.folder.description,
            'date_modified': self.get_drf_datetime(self.folder.date_modified),
            'sodar_uuid': str(self.folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_get_invalid_uuid(self):
        """Test GET with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_retrieve_update_destroy',
                kwargs={'folder': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)

    def test_put_update(self):
        """Test PUT to update folder"""
        folder_data = {
            'name': 'UPDATED Folder',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
        }
        response = self.request_knox(self.url, method='PUT', data=folder_data)
        self.assertEqual(response.status_code, 200, msg=response.data)
        self.folder.refresh_from_db()
        expected = {
            **folder_data,
            'folder': None,
            'owner': str(self.folder.owner.sodar_uuid),
            'project': str(self.folder.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(self.folder.date_modified),
            'sodar_uuid': str(self.folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_delete(self):
        """Test DELETE to remove folder"""
        response = self.request_knox(self.url, method='DELETE')
        self.assertEqual(response.status_code, 204, msg=response.data)
        self.assertIsNone(response.data)
        with self.assertRaises(Folder.DoesNotExist):
            Folder.objects.get(
                project=self.project, sodar_uuid=self.folder.sodar_uuid
            )


class TestFileListCreateAPIView(FilesfoldersAPIViewTestBase):
    """Tests for FileListCreateAPIView"""

    def setUp(self):
        super().setUp()
        self.file_data = {
            'name': 'New File',
            'flag': 'IMPORTANT',
            'description': 'File\'s description',
            'secret': 'foo',
            'public_url': True,
            'file': open(ZIP_PATH_NO_FILES, 'rb'),
        }
        self.url = reverse(
            'filesfolders:api_file_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )

    def tearDown(self):
        self.file_data['file'].close()
        super().tearDown()

    def test_get_list(self):
        """Test FileListCreateAPIView GET to list files"""
        response = self.request_knox(self.url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'name': self.file.name,
            'folder': None,
            'owner': str(self.file.owner.sodar_uuid),
            'project': str(self.file.project.sodar_uuid),
            'flag': self.file.flag,
            'description': self.file.description,
            'secret': self.file.secret,
            'public_url': self.file.public_url,
            'date_modified': self.get_drf_datetime(self.file.date_modified),
            'sodar_uuid': str(self.file.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), [expected])

    def test_get_pagination(self):
        """Test GET with pagination"""
        url = self.url + '?page=1'
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'count': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'name': self.file.name,
                    'folder': None,
                    'owner': str(self.file.owner.sodar_uuid),
                    'project': str(self.file.project.sodar_uuid),
                    'flag': self.file.flag,
                    'description': self.file.description,
                    'secret': self.file.secret,
                    'public_url': self.file.public_url,
                    'date_modified': self.get_drf_datetime(
                        self.file.date_modified
                    ),
                    'sodar_uuid': str(self.file.sodar_uuid),
                }
            ],
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_post_create_root(self):
        """Test POST to create file in root"""
        response = self.request_knox(
            self.url, method='POST', format='multipart', data=self.file_data
        )
        self.assertEqual(response.status_code, 201, msg=response.data)
        new_file = File.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_file)
        self.assertNotEqual(new_file.file.file.size, 0)
        expected = {
            **self.file_data,
            'folder': None,
            'owner': str(self.user.sodar_uuid),
            'project': str(self.project.sodar_uuid),
            'secret': new_file.secret,
            'public_url': new_file.public_url,
            'date_modified': self.get_drf_datetime(new_file.date_modified),
            'sodar_uuid': str(new_file.sodar_uuid),
        }
        expected.pop('file')
        self.assertEqual(json.loads(response.content), expected)

    def test_post_create_folder(self):
        """Test POST to create file under folder"""
        file_data = {**self.file_data, 'folder': str(self.folder.sodar_uuid)}
        response = self.request_knox(
            self.url, method='POST', format='multipart', data=file_data
        )
        self.assertEqual(response.status_code, 201, msg=response.data)
        new_file = File.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_file)
        self.assertNotEqual(new_file.file.file.size, 0)
        expected = {
            **file_data,
            'owner': str(self.user.sodar_uuid),
            'project': str(self.project.sodar_uuid),
            'public_url': self.file_data['public_url'],
            'secret': new_file.secret,
            'date_modified': self.get_drf_datetime(new_file.date_modified),
            'sodar_uuid': str(new_file.sodar_uuid),
        }
        expected.pop('file')
        self.assertEqual(json.loads(response.content), expected)

    def test_post_create_category(self):
        """Test POST to create file in category (should fail)"""
        category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.make_assignment(category, self.user, self.role_owner)
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_list_create',
                kwargs={'project': category.sodar_uuid},
            ),
            method='POST',
            format='multipart',
            data=self.file_data,
        )
        self.assertEqual(response.status_code, 403, msg=response.data)
        self.assertEqual(
            str(response.data['detail']),
            INVALID_PROJECT_TYPE_MSG.format(project_type=PROJECT_TYPE_CATEGORY),
        )


class TestFileRetrieveUpdateDestroyAPIView(FilesfoldersAPIViewTestBase):
    """Tests for FileRetrieveUpdateDestroyAPIView"""

    def setUp(self):
        super().setUp()
        self.file_data = {
            'name': 'UPDATED File',
            'flag': 'FLAG',
            'description': 'UPDATED description',
            'secret': 'foo',
            'public_url': False,
            'file': open(ZIP_PATH_NO_FILES, 'rb'),
        }
        self.url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': self.file.sodar_uuid},
        )

    def tearDown(self):
        self.file_data['file'].close()
        super().tearDown()

    def test_get_retrieve(self):
        """Test FileRetrieveUpdateDestroyAPIView GET to retrieve file"""
        response = self.request_knox(self.url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'name': self.file.name,
            'folder': None,
            'owner': str(self.file.owner.sodar_uuid),
            'project': str(self.file.project.sodar_uuid),
            'flag': self.file.flag,
            'description': self.file.description,
            'public_url': self.file.public_url,
            'secret': self.file.secret,
            'date_modified': self.get_drf_datetime(self.file.date_modified),
            'sodar_uuid': str(self.file.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_get_invalid_uuid(self):
        """Test GET with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_retrieve_update_destroy',
                kwargs={'file': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)

    def test_put_update(self):
        """Test PUT to update file"""
        response = self.request_knox(
            self.url, method='PUT', format='multipart', data=self.file_data
        )
        self.assertEqual(response.status_code, 200, msg=response.data)
        old_secret = self.file.secret
        self.file.refresh_from_db()
        self.assertEqual(self.file.name, self.file_data['name'])
        self.assertEqual(self.file.flag, self.file_data['flag'])
        self.assertEqual(self.file.description, self.file_data['description'])
        self.assertNotEqual(
            self.file.secret,
            old_secret,
            msg='Secret should change when public_url flag changes',
        )
        expected = {
            **self.file_data,
            'folder': None,
            'owner': str(self.file.owner.sodar_uuid),
            'project': str(self.file.project.sodar_uuid),
            'public_url': self.file_data['public_url'],
            'secret': self.file.secret,
            'date_modified': self.get_drf_datetime(self.file.date_modified),
            'sodar_uuid': str(self.file.sodar_uuid),
        }
        expected.pop('file')
        self.assertEqual(json.loads(response.content), expected)

    def test_delete(self):
        """Test DELETE"""
        response = self.request_knox(self.url, method='DELETE')
        self.assertEqual(response.status_code, 204, msg=response.data)
        self.assertIsNone(response.data)
        with self.assertRaises(File.DoesNotExist):
            File.objects.get(
                project=self.project, sodar_uuid=self.file.sodar_uuid
            )


class TestFileServeAPIView(FilesfoldersAPIViewTestBase):
    """Tests for FileServeAPIView"""

    def test_get(self):
        """Test FileServeAPIView GET to download file"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_serve',
                kwargs={'file': self.file.sodar_uuid},
            )
        )
        expected = b'content'
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(response.content, expected)

    def test_get_not_found(self):
        """Test GET with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_serve',
                kwargs={'file': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)


class TestHyperLinkListCreateAPIView(FilesfoldersAPIViewTestBase):
    """Tests for HyperLinkListCreateAPIView"""

    def setUp(self):
        super().setUp()
        self.hyperlink_data = {
            'name': 'New HyperLink',
            'flag': 'IMPORTANT',
            'description': 'HyperLink\'s description',
            'url': 'http://www.cubi.bihealth.org',
        }
        self.url = reverse(
            'filesfolders:api_hyperlink_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_get_list(self):
        """Test HyperLinkListCreateAPIView GET to list hyperlinks"""
        response = self.request_knox(self.url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'name': self.hyperlink.name,
            'folder': None,
            'owner': str(self.hyperlink.owner.sodar_uuid),
            'project': str(self.hyperlink.project.sodar_uuid),
            'flag': self.hyperlink.flag,
            'description': self.hyperlink.description,
            'url': self.hyperlink.url,
            'date_modified': self.get_drf_datetime(
                self.hyperlink.date_modified
            ),
            'sodar_uuid': str(self.hyperlink.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), [expected])

    def test_get_pagination(self):
        """Test GET with pagination"""
        url = self.url + '?page=1'
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'count': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'name': self.hyperlink.name,
                    'folder': None,
                    'owner': str(self.hyperlink.owner.sodar_uuid),
                    'project': str(self.hyperlink.project.sodar_uuid),
                    'flag': self.hyperlink.flag,
                    'description': self.hyperlink.description,
                    'url': self.hyperlink.url,
                    'date_modified': self.get_drf_datetime(
                        self.hyperlink.date_modified
                    ),
                    'sodar_uuid': str(self.hyperlink.sodar_uuid),
                }
            ],
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_post_root(self):
        """Test POST to create hyperlink in root"""
        response = self.request_knox(
            self.url, method='POST', data=self.hyperlink_data
        )
        self.assertEqual(response.status_code, 201, msg=response.data)
        new_link = HyperLink.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_link)
        expected = {
            **self.hyperlink_data,
            'folder': None,
            'owner': str(self.user.sodar_uuid),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_link.date_modified),
            'sodar_uuid': str(new_link.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_post_folder(self):
        """Test POST to create hyperlink under folder"""
        hyperlink_data = {
            **self.hyperlink_data,
            'folder': str(self.folder.sodar_uuid),
        }
        response = self.request_knox(
            self.url, method='POST', data=hyperlink_data
        )
        self.assertEqual(response.status_code, 201, msg=response.data)
        new_link = HyperLink.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_link)
        expected = {
            **hyperlink_data,
            'owner': str(self.user.sodar_uuid),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_link.date_modified),
            'sodar_uuid': str(new_link.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_post_category(self):
        """Test POST to create hyperlink in category (should fail)"""
        category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.make_assignment(category, self.user, self.role_owner)
        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_list_create',
                kwargs={'project': category.sodar_uuid},
            ),
            method='POST',
            data=self.hyperlink_data,
        )
        self.assertEqual(response.status_code, 403, msg=response.data)
        self.assertEqual(
            str(response.data['detail']),
            INVALID_PROJECT_TYPE_MSG.format(project_type=PROJECT_TYPE_CATEGORY),
        )


class TestHyperLinkRetrieveUpdateDestroyAPIView(FilesfoldersAPIViewTestBase):
    """Tests for HyperLinkRetrieveUpdateDestroyAPIView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': self.hyperlink.sodar_uuid},
        )

    def test_get_retrieve(self):
        """Test HyperLinkRetrieveUpdateDestroyAPIView to retrieve hyperlink"""
        response = self.request_knox(self.url)
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'name': self.hyperlink.name,
            'folder': None,
            'owner': str(self.hyperlink.owner.sodar_uuid),
            'project': str(self.hyperlink.project.sodar_uuid),
            'flag': self.hyperlink.flag,
            'description': self.hyperlink.description,
            'url': self.hyperlink.url,
            'date_modified': self.get_drf_datetime(
                self.hyperlink.date_modified
            ),
            'sodar_uuid': str(self.hyperlink.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_get_invalid_uuid(self):
        """Test GET with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_retrieve_update_destroy',
                kwargs={'hyperlink': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)

    def test_put_update(self):
        """Test PUT to update hyperlink"""
        hyperlink_data = {
            'name': 'UPDATED HyperLink',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
            'url': 'http://www.bihealth.org',
        }
        response = self.request_knox(
            self.url, method='PUT', data=hyperlink_data
        )
        self.assertEqual(response.status_code, 200, msg=response.data)
        self.hyperlink.refresh_from_db()
        self.assertEqual(self.hyperlink.name, hyperlink_data['name'])
        self.assertEqual(self.hyperlink.flag, hyperlink_data['flag'])
        self.assertEqual(
            self.hyperlink.description, hyperlink_data['description']
        )
        expected = {
            **hyperlink_data,
            'folder': None,
            'owner': str(self.hyperlink.owner.sodar_uuid),
            'project': str(self.hyperlink.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(
                self.hyperlink.date_modified
            ),
            'sodar_uuid': str(self.hyperlink.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_delete(self):
        """Test DELETE"""
        response = self.request_knox(self.url, method='DELETE')
        self.assertEqual(response.status_code, 204, msg=response.data)
        self.assertIsNone(response.data)



# Generated by Django 4.2.19 on 2025-03-05 15:08

from django.db import migrations


FIX_VALUES = [('True', '1'), ('False', '0')]


def fix_project_star(apps, schema_editor):
    """
    Fix project_star app setting values. Due to an earlier bug, settings were
    set with values "True" and "False" instead of "1" and "0".
    """
    AppSetting = apps.get_model('projectroles', 'AppSetting')
    for v in FIX_VALUES:
        for a in AppSetting.objects.filter(name='project_star', value=v[0]):
            a.value = v[1]
            a.save()


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0033_alter_sodaruseradditionalemail_secret'),
    ]

    operations = [
        migrations.RunPython(
            code=fix_project_star,
            reverse_code=migrations.RunPython.noop,
        )
    ]



"""Utility functions for the filesfolders app"""

from django.http import HttpRequest
from django.urls import reverse

from filesfolders.models import File


def build_public_url(file: File, request: HttpRequest) -> str:
    """
    Return public URL for a file.

    :param file: File object for which public URL will be created
    :param request: HTTP request of View calling the function
    :return: URL (string)
    """
    return request.build_absolute_uri(
        reverse(
            'filesfolders:file_serve_public',
            kwargs={'secret': file.secret, 'file_name': file.name},
        )
    )



"""Timeline API for adding and updating events"""

import logging
import re

from typing import Any, Optional

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import QuerySet
from django.http import HttpRequest
from django.urls import reverse
from django.utils.text import Truncator

# Projectroles dependency
from projectroles.models import Project, RemoteSite, SODARUser
from projectroles.plugins import PluginAPI
from projectroles.templatetags.projectroles_common_tags import get_user_html
from projectroles.utils import get_app_names

from timeline.models import (
    TimelineEvent,
    TimelineEventObjectRef,
    EVENT_STATUS_TYPES,
    OBJ_REF_UNNAMED,
    TL_STATUS_OK,
    TL_STATUS_INIT,
    TL_STATUS_SUBMIT,
    TL_STATUS_FAILED,
    TL_STATUS_INFO,
    TL_STATUS_CANCEL,
)


logger = logging.getLogger(__name__)
plugin_api = PluginAPI()
User = get_user_model()


# Local variables
APP_NAMES = get_app_names()
LABEL_MAX_WIDTH = 32
UNKNOWN_LABEL = '(unknown)'
PLUGIN_NOT_FOUND_MSG = 'Plugin not found: {plugin_name}'


class TimelineAPI:
    """Timeline backend API to be used by Django apps."""

    # Attributes --------------------------------------------------------------

    TL_STATUS_OK = TL_STATUS_OK
    TL_STATUS_INIT = TL_STATUS_INIT
    TL_STATUS_SUBMIT = TL_STATUS_SUBMIT
    TL_STATUS_FAILED = TL_STATUS_FAILED
    TL_STATUS_INFO = TL_STATUS_INFO
    TL_STATUS_CANCEL = TL_STATUS_CANCEL

    # Internal Helpers ---------------------------------------------------------

    @classmethod
    def _get_ref_label(cls, label: str) -> str:
        """Return reference object name in displayable form"""
        if not label:
            return OBJ_REF_UNNAMED
        if not {' ', '-'}.intersection(label):
            return Truncator(label).chars(LABEL_MAX_WIDTH)
        return label

    @classmethod
    def _get_history_link(cls, obj_ref: TimelineEventObjectRef) -> str:
        """Return history link HTML for event object reference"""
        url_name = 'timeline:list_object_site'
        url_kwargs = {
            'object_model': obj_ref.object_model,
            'object_uuid': obj_ref.object_uuid,
        }
        if obj_ref.event.project:
            url_name = 'timeline:list_object'
            url_kwargs['project'] = obj_ref.event.project.sodar_uuid
        history_url = reverse(url_name, kwargs=url_kwargs)
        return (
            f'<a href="{history_url}" class="sodar-tl-object-link">'
            f'<i class="iconify" '
            f'data-icon="mdi:clock-time-eight-outline"></i></a>'
        )

    @classmethod
    def _get_not_found_label(cls, obj_ref: TimelineEventObjectRef) -> str:
        """Get label for object which is not found in the database"""
        return (
            f'<span class="text-danger">{cls._get_ref_label(obj_ref.name)}'
            f'</span> {cls._get_history_link(obj_ref)}'
        )

    @classmethod
    def _get_project_desc(
        cls,
        obj_ref: TimelineEventObjectRef,
        request: Optional[HttpRequest] = None,
    ) -> str:
        """Get description HTML for special case: Project model"""
        project = Project.objects.filter(sodar_uuid=obj_ref.object_uuid).first()
        if (
            project
            and request
            and request.user.has_perm('projectroles.view_project', project)
        ):
            return '<a class="sodar-tl-project-link" href="{}">{}</a>'.format(
                reverse(
                    'projectroles:detail',
                    kwargs={'project': project.sodar_uuid},
                ),
                cls._get_ref_label(project.title),
            )
        elif project:
            return '<span class="text-danger">{}</span>'.format(
                cls._get_ref_label(project.title)
            )
        return obj_ref.name

    @classmethod
    def _get_remote_site_desc(
        cls,
        obj_ref: TimelineEventObjectRef,
        request: Optional[HttpRequest] = None,
    ) -> str:
        """Get description HTML for special case: RemoteSite model"""
        site = RemoteSite.objects.filter(sodar_uuid=obj_ref.object_uuid).first()
        if site and request and request.user.is_superuser:
            return '<a href="{}">{}</a> {}'.format(
                reverse(
                    'projectroles:remote_projects',
                    kwargs={'remotesite': site.sodar_uuid},
                ),
                site.name,
                cls._get_history_link(obj_ref),
            )
        elif site:
            return site.name
        return cls._get_not_found_label(obj_ref)

    @classmethod
    def _get_ref_description(
        cls,
        event: TimelineEvent,
        ref_label: str,
        app_plugin: Any,
        request: Optional[HttpRequest] = None,
    ) -> str:
        """
        Get reference object description for event description, or unknown label
        if not found.

        :param event: TimelineEvent object
        :param ref_label: Label for the reference object (string)
        :param app_plugin: App plugin or None
        :param request: Request object or None
        :return: String (contains HTML)
        """
        # Special case: Extra data reference
        if ref_label.startswith('extra-'):
            desc = app_plugin.get_extra_data_link(event.extra_data, ref_label)
            return desc if desc else UNKNOWN_LABEL

        # Get object reference
        try:
            obj_ref = TimelineEventObjectRef.objects.get(
                event=event, label=ref_label
            )
        except TimelineEventObjectRef.DoesNotExist:
            return UNKNOWN_LABEL

        # Special case: User model
        if obj_ref.object_model == 'User':
            try:
                user = User.objects.get(sodar_uuid=obj_ref.object_uuid)
                return f'{get_user_html(user)} {cls._get_history_link(obj_ref)}'
            except User.DoesNotExist:
                return UNKNOWN_LABEL

        # Special case: Project model
        elif obj_ref.object_model == 'Project':
            return cls._get_project_desc(obj_ref, request)

        # Special case: RemoteSite model
        elif obj_ref.object_model == 'RemoteSite':
            return cls._get_remote_site_desc(obj_ref, request)

        # Special case: projectroles app
        elif event.app == 'projectroles':
            return cls._get_not_found_label(obj_ref)

        # Apps with plugins
        else:
            try:
                link = app_plugin.get_object_link(
                    obj_ref.object_model, obj_ref.object_uuid
                )
            except Exception as ex:
                logger.error(
                    f'Exception in {app_plugin.name}.get_object_link(): {ex}'
                )
                if settings.DEBUG:
                    raise ex
                link = None
            if link:
                if not link.name:
                    logger.warning(
                        f'Empty name returned by plugin "{app_plugin.name}" '
                        f'for object reference "{obj_ref}" '
                        f'({obj_ref.object_uuid})"'
                    )
                return '<a href="{}" {}>{}</a> {}'.format(
                    link.url,
                    'target="_blank"' if link.blank else '',
                    cls._get_ref_label(link.name),
                    cls._get_history_link(obj_ref),
                )
            else:
                return cls._get_not_found_label(obj_ref)

    # API functions ------------------------------------------------------------

    @classmethod
    def add_event(
        cls,
        project: Project,
        app_name: str,
        user: Optional[SODARUser],
        event_name: str,
        description: str,
        classified: bool = False,
        extra_data: dict = None,
        status_type: str = None,
        status_desc: str = None,
        status_extra_data: dict = None,
        plugin_name: str = None,
    ) -> TimelineEvent:
        """
        Create and save a timeline event.

        :param project: Project object or None
        :param app_name: Name of app from which event was invoked (must
            correspond to "name" attribute of app plugin)
        :param user: User invoking the event or None
        :param event_name: Event ID string (must match schema)
        :param description: Description of status change (may include {object
            label} references)
        :param classified: Whether event is classified (boolean, optional)
        :param extra_data: Additional event data (dict, optional)
        :param status_type: Initial status type (string, optional)
        :param status_desc: Initial status description (string, optional)
        :param status_extra_data: Extra data for initial status (dict, optional)
        :param plugin_name: Name of plugin to which the event is related
            (optional, plugin with the name of the app is assumed if unset)
        :return: TimelineEvent object
        :raise: ValueError if app_name or status_type is invalid
        """
        if app_name not in APP_NAMES:
            raise ValueError(
                f'Unknown app name "{app_name}" (active apps: '
                f'{", ".join(x for x in APP_NAMES)})'
            )
        if status_type and status_type not in EVENT_STATUS_TYPES:
            raise ValueError(
                f'Unknown status type "{status_type}" (valid types: '
                f'{", ".join(x for x in EVENT_STATUS_TYPES)})'
            )

        # Handle user in case called with AnonymousUser object
        if user and user.is_anonymous:
            user = None

        event = TimelineEvent()
        event.project = project
        event.app = app_name
        event.plugin = plugin_name
        event.user = user
        event.event_name = event_name
        event.description = description
        event.classified = classified
        if extra_data:
            event.extra_data = extra_data
        event.save()

        # Always add "INIT" status when creating, except for "INFO"
        if status_type not in [TL_STATUS_INFO, TL_STATUS_INIT]:
            event.set_status(TL_STATUS_INIT)
        # Add additional status if set (use if e.g. event is immediately "OK")
        if status_type:
            event.set_status(status_type, status_desc, status_extra_data)
        return event

    @classmethod
    def get_project_events(
        cls, project: Project, classified: bool = False
    ) -> QuerySet:
        """
        Return timeline events for a project.

        :param project: Project object
        :param classified: Include classified (boolean)
        :return: QuerySet
        """
        events = TimelineEvent.objects.filter(project=project)
        if not classified:
            events = events.filter(classified=False)
        return events

    @classmethod
    def get_event_description(
        cls,
        event: TimelineEvent,
        plugin_lookup: Optional[dict] = None,
        request: Optional[HttpRequest] = None,
    ) -> str:
        """
        Return the description of a timeline event as HTML.

        :param event: TimelineEvent object
        :param plugin_lookup: App plugin lookup dict (optional)
        :param request: Request object (optional)
        :return: String (contains HTML)
        """
        desc = event.description
        ref_ids = re.findall('{\'?(.*?)\'?}', desc)
        if len(ref_ids) == 0:
            return event.description
        refs = {}
        app_plugin = None

        if event.app != 'projectroles':
            plugin_name = event.plugin if event.plugin else event.app
            if plugin_lookup:
                app_plugin = plugin_lookup.get(plugin_name)
            else:
                app_plugin = plugin_api.get_app_plugin(plugin_name)
            if not app_plugin:
                msg = PLUGIN_NOT_FOUND_MSG.format(plugin_name=plugin_name)
                logger.error(msg + f' (UUID={event.sodar_uuid})')
                return (
                    f'<span class="sodar-tl-plugin-error text-danger">{msg}'
                    f'</span>'
                )

        # Get links for object references
        for r in ref_ids:
            refs[r] = cls._get_ref_description(event, r, app_plugin, request)
        try:
            return event.description.format(**refs)
        except Exception as ex:  # Dispaly exception instead of crashing
            logger.error(
                f'Error formatting event description: {ex} '
                f'(UUID={event.sodar_uuid})'
            )
            return (
                f'<span class="sodar-tl-format-error text-danger">'
                f'{ex.__class__.__name__}: {ex}</span>'
            )

    @classmethod
    def get_object_url(cls, obj: Any, project: Optional[Project] = None) -> str:
        """
        Return the URL for a timeline event object history.

        :param obj: Django database object
        :param project: Related Project object or None
        :return: String
        """
        url_name = 'timeline:list_object_site'
        url_kwargs = {
            'object_model': obj.__class__.__name__,
            'object_uuid': obj.sodar_uuid,
        }
        if project:
            url_name = 'timeline:list_object'
            url_kwargs['project'] = project.sodar_uuid
        return reverse(url_name, kwargs=url_kwargs)

    @classmethod
    def get_object_link(
        cls, obj: Any, project: Optional[Project] = None
    ) -> str:
        """
        Return an inline HTML icon link for a timeline event object history.

        :param obj: Django database object
        :param project: Related Project object or None
        :return: String (contains HTML)
        """
        return (
            f'<a href="{cls.get_object_url(obj, project)}" '
            f'class="sodar-tl-object-link"><i class="iconify" '
            f'data-icon="mdi:clock-time-eight-outline"></i></a>'
        )

    @classmethod
    def get_models(
        cls,
    ) -> tuple[type[TimelineEvent], type[TimelineEventObjectRef]]:
        """
        Return timeline event model classes for custom/advanced queries.

        :return: TimelineEvent, TimelineEventObjectRef
        """
        return TimelineEvent, TimelineEventObjectRef



"""Tests for REST API view permissions in the filesfolders app"""

import os
import uuid

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions_api import ProjectAPIPermissionTestBase

from filesfolders.models import File, Folder, HyperLink
from filesfolders.tests.test_permissions import FilesfoldersPermissionTestMixin
from filesfolders.views_api import (
    FILESFOLDERS_API_MEDIA_TYPE,
    FILESFOLDERS_API_DEFAULT_VERSION,
)


# Local constants
SECRET = '7dqq83clo2iyhg29hifbor56og6911r5'
TEST_DATA_PATH = os.path.dirname(__file__) + '/data/'
ZIP_PATH_NO_FILES = TEST_DATA_PATH + 'no_files.zip'
OBJ_UUID = uuid.uuid4()


class FilesfoldersAPIPermissionTestBase(
    FilesfoldersPermissionTestMixin, ProjectAPIPermissionTestBase
):
    """Base class for filesfolders REST API view permission tests"""

    media_type = FILESFOLDERS_API_MEDIA_TYPE
    api_version = FILESFOLDERS_API_DEFAULT_VERSION


class TestFolderListCreateAPIView(FilesfoldersAPIPermissionTestBase):
    """Tests for FolderListCreateAPIView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:api_folder_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.post_data = {
            'name': 'New Folder',
            'flag': 'IMPORTANT',
            'description': 'Description',
        }
        self.good_users_get = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users_get = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test FolderListCreateAPIView GET"""
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_post(self):
        """Test FolderListCreateAPIView POST"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.assert_response_api(
            self.url, good_users, 201, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, bad_users, 403, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            good_users,
            201,
            method='POST',
            data=self.post_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            data=self.post_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_anon(self):
        """Test POST with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )

    def test_post_archive(self):
        """Test POST with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url, self.superuser, 201, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            data=self.post_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            data=self.post_data,
        )

    def test_post_block(self):
        """Test POST with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url, self.superuser, 201, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )

    def test_post_read_only(self):
        """Test POST with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url, self.superuser, 201, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )


class TestFolderRetrieveUpdateDestroyAPIView(FilesfoldersAPIPermissionTestBase):
    """Tests for FolderRetrieveUpdateDestroyAPIView permissions"""

    def _make_folder(self) -> Folder:
        folder = self.make_test_folder()
        folder.sodar_uuid = OBJ_UUID
        folder.save()
        return folder

    def setUp(self):
        super().setUp()
        folder = self._make_folder()
        self.url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': folder.sodar_uuid},
        )
        self.put_data = {
            'name': 'UPDATED Folder',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
        }
        self.patch_data = {'name': 'UPDATED Folder'}
        self.good_users_get = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users_get = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.good_users_update = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        self.bad_users_update = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test FolderRetrieveUpdateDestroyAPIView GET"""
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_put(self):
        """Test FolderRetrieveUpdateDestroyAPIView PUT"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url,
            self.bad_users_update,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PUT',
            data=self.put_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_put_anon(self):
        """Test PUT with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_put_archive(self):
        """Test PUT with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_put_block(self):
        """Test PUT with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url, self.superuser, 200, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_put_read_only(self):
        """Test PUT with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_patch(self):
        """Test FolderRetrieveUpdateDestroyAPIView PATCH"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.bad_users_update,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PATCH',
            data=self.patch_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            data=self.patch_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_patch_anon(self):
        """Test PATCH for with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )

    def test_patch_archive(self):
        """Test PATCH with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            data=self.patch_data,
        )

    def test_patch_block(self):
        """Test PATCH with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url, self.superuser, 200, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )

    def test_patch_read_only(self):
        """Test PATCH with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )

    def test_delete(self):
        """Test FolderRetrieveUpdateDestroyAPIView DELETE"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            204,
            method='DELETE',
            cleanup_method=self._make_folder,
        )
        self.assert_response_api(
            self.url, self.bad_users_update, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')
        self.assert_response_api(
            self.url,
            self.good_users_update,
            204,
            method='DELETE',
            cleanup_method=self._make_folder,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='DELETE'
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            data=self.patch_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_delete_anon(self):
        """Test DELETE with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_archive(self):
        """Test DELETE with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._make_folder,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='DELETE'
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            data=self.patch_data,
        )

    def test_delete_block(self):
        """Test DELETE with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._make_folder,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_read_only(self):
        """Test DELETE with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._make_folder,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')


class TestFileListCreateAPIView(FilesfoldersAPIPermissionTestBase):
    """Tests for FileListCreateAPIView permissions"""

    def _make_post_data(self) -> dict:
        self.post_data = {
            'name': 'New File',
            'flag': 'IMPORTANT',
            'description': 'File\'s description',
            'secret': 'foo',
            'public_url': True,
            'file': open(ZIP_PATH_NO_FILES, 'rb'),
        }

    def _cleanup(self):
        File.objects.all().delete()
        if hasattr(self, 'post_data'):
            self.post_data['file'].seek(0)

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:api_file_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.good_users_get = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users_get = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def tearDown(self):
        if hasattr(self, 'post_data'):
            self.post_data['file'].close()
        super().tearDown()

    def test_get(self):
        """Test FileListCreateAPIView GET"""
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_post(self):
        """Test FileListCreateAPIView POST"""
        self._make_post_data()
        # NOTE: Must call cleanup for ALL requests to seek the file
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.assert_response_api(
            self.url,
            good_users,
            201,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            bad_users,
            403,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            good_users,
            201,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_anon(self):
        """Test POST with anonymous access"""
        self._make_post_data()
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )

    def test_post_archive(self):
        """Test POST with archived project"""
        self._make_post_data()
        self.project.set_archive()
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )

    def test_post_block(self):
        """Test POST with project access block"""
        self.set_access_block(self.project)
        self._make_post_data()
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )

    def test_post_read_only(self):
        """Test POST with site read-only mode"""
        self.set_site_read_only()
        self._make_post_data()
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )


class TestFileRetrieveUpdateDestroyAPIView(FilesfoldersAPIPermissionTestBase):
    """Tests for FileRetrieveUpdateDestroyAPIView permissions"""

    def _make_file(self) -> File:
        file = self.make_test_file()
        file.sodar_uuid = OBJ_UUID
        file.save()
        return file

    def _make_put_data(self):
        self.put_data = {
            'name': 'UPDATED File',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
            'secret': 'foo',
            'public_url': True,
            'file': open(ZIP_PATH_NO_FILES, 'rb'),
        }

    def _cleanup_put(self):
        self.put_data['file'].seek(0)

    def setUp(self):
        super().setUp()
        file = self._make_file()
        self.patch_data = {'name': 'UPDATED File'}
        self.url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': file.sodar_uuid},
        )
        self.good_users_get = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users_get = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.good_users_update = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        self.bad_users_update = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test FileRetrieveUpdateDestroyAPIView GET"""
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_put(self):
        """Test FileRetrieveUpdateDestroyAPIView PUT"""
        self._make_put_data()
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.bad_users_update,
            403,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PUT',
            format='multipart',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.put_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_put_anon(self):
        """Test PUT with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self._make_put_data()
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.put_data,
        )

    def test_put_archive(self):
        """Test PUT with archived project"""
        self.project.set_archive()
        self._make_put_data()
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PUT',
            format='multipart',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.put_data,
        )

    def test_put_block(self):
        """Test PUT with project access block"""
        self.set_access_block(self.project)
        self._make_put_data()
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )

    def test_put_read_only(self):
        """Test PUT with site read-only mode"""
        self.set_site_read_only()
        self._make_put_data()
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.put_data,
            cleanup_method=self._cleanup_put,
        )

    def test_patch(self):
        """Test FileRetrieveUpdateDestroyAPIView PATCH"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.bad_users_update,
            403,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_patch_anon(self):
        """Test PATCH with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )

    def test_patch_archive(self):
        """Test PATCH with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        # Test public project
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )

    def test_patch_block(self):
        """Test PATCH with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )

    def test_patch_read_only(self):
        """Test PATCH with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.patch_data,
        )

    def test_delete(self):
        """Test FileRetrieveUpdateDestroyAPIView DELETE"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            204,
            method='DELETE',
            cleanup_method=self._make_file,
        )
        self.assert_response_api(
            self.url, self.bad_users_update, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')
        self.assert_response_api(
            self.url,
            self.good_users_update,
            204,
            method='DELETE',
            cleanup_method=self._make_file,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='DELETE',
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='DELETE',
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_delete_anon(self):
        """Test DELETE with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_archive(self):
        """Test DELETE with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._make_file,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_block(self):
        """Test DELETE with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._make_file,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_read_only(self):
        """Test DELETE with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._make_file,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')


class TestFileServeAPIView(FilesfoldersAPIPermissionTestBase):
    """Tests for FileServeAPIView permissions"""

    def setUp(self):
        super().setUp()
        file = self.make_test_file()
        self.url = reverse(
            'filesfolders:api_file_serve', kwargs={'file': file.sodar_uuid}
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test FileServeAPIView GET"""
        self.assert_response_api(self.url, self.good_users, 200, method='GET')
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users, 200, method='GET')
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users, 200, method='GET')
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)


class TestHyperLinkListCreateAPIView(FilesfoldersAPIPermissionTestBase):
    """Tests for HyperLinkListCreateAPIView permissions"""

    @classmethod
    def _cleanup(cls):
        HyperLink.objects.all().delete()

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:api_hyperlink_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.post_data = {
            'name': 'New HyperLink',
            'flag': 'IMPORTANT',
            'description': 'Description',
            'url': 'https://www.cubi.bihealth.org',
        }
        self.good_users_get = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users_get = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test HyperLinkListCreateAPIView GET"""
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_post(self):
        """Test HyperLinkListCreateAPIView POST"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.assert_response_api(
            self.url,
            good_users,
            201,
            method='POST',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url, bad_users, 403, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            good_users,
            201,
            method='POST',
            data=self.post_data,
            cleanup_method=self._cleanup,
            knox=True,
        )
        # Test public project
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            data=self.post_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_anon(self):
        """Test POST with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )

    def test_post_archive(self):
        """Test POST with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            data=self.post_data,
            cleanup_method=self._cleanup,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='POST',
            data=self.post_data,
        )

    def test_post_block(self):
        """Test POST with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )

    def test_post_read_only(self):
        """Test POST with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url,
            self.superuser,
            201,
            method='POST',
            data=self.post_data,
            cleanup_method=self._cleanup,
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='POST',
            data=self.post_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )


class TestHyperLinkRetrieveUpdateDestroyAPIView(
    FilesfoldersAPIPermissionTestBase
):
    """Tests for HyperLinkRetrieveUpdateDestroyAPIView permissions"""

    def _make_link(self) -> HyperLink:
        link = self.make_test_link()
        link.sodar_uuid = OBJ_UUID
        link.save()
        return link

    def _cleanup_delete(self):
        self._make_link()

    def setUp(self):
        super().setUp()
        link = self._make_link()
        self.url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': link.sodar_uuid},
        )
        self.put_data = {
            'name': 'UPDATED HyperLink',
            'flag': 'FLAG',
            'description': 'UPDATED description',
            'url': 'https://www.bihealth.org',
        }
        self.patch_data = {'name': 'UPDATED Hyperlink'}
        self.good_users_get = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users_get = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.good_users_update = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,  # Owner of link
            self.user_delegate,
        ]
        self.bad_users_update = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test HyperLinkRetrieveUpdateDestroyAPIView GET"""
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users_get, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users_get, 200)
        self.assert_response_api(self.url, self.bad_users_get, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_put(self):
        """Test HyperLinkRetrieveUpdateDestroyAPIView PUT"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url,
            self.bad_users_update,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PUT',
            data=self.put_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_put_anon(self):
        """Test PUT with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_put_archive(self):
        """Test PUT with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PUT',
            data=self.put_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_put_block(self):
        """Test PUT with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url, self.superuser, 200, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_put_read_only(self):
        """Test PUT with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PUT', data=self.put_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PUT',
            data=self.put_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PUT', data=self.put_data
        )

    def test_patch(self):
        """Test HyperLinkRetrieveUpdateDestroyAPIView PATCH"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.bad_users_update,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.good_users_update,
            200,
            method='PATCH',
            data=self.patch_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            data=self.patch_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_patch_anon(self):
        """Test PATCH with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )

    def test_patch_archive(self):
        """Test PATCH with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.superuser,
            200,
            method='PATCH',
            data=self.patch_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url,
            self.anonymous,
            401,
            method='PATCH',
            data=self.patch_data,
        )

    def test_patch_block(self):
        """Test PATCH with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url, self.superuser, 200, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )

    def test_patch_read_only(self):
        """Test PATCH with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url, self.superuser, 200, method='PATCH', data=self.patch_data
        )
        self.assert_response_api(
            self.url,
            self.auth_non_superusers,
            403,
            method='PATCH',
            data=self.patch_data,
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='PATCH', data=self.patch_data
        )

    def test_delete(self):
        """Test HyperLinkRetrieveUpdateDestroyAPIView DELETE"""
        self.assert_response_api(
            self.url,
            self.good_users_update,
            204,
            method='DELETE',
            cleanup_method=self._cleanup_delete,
        )
        self.assert_response_api(
            self.url, self.bad_users_update, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')
        self.assert_response_api(
            self.url,
            self.good_users_update,
            204,
            method='DELETE',
            cleanup_method=self._cleanup_delete,
            knox=True,
        )
        # Test public project
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='DELETE',
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_delete_anon(self):
        """Test DELETE with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_archive(self):
        """Test DELETE with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._cleanup_delete,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._cleanup_delete,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.user_no_roles, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_block(self):
        """Test DELETE with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._cleanup_delete,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')

    def test_delete_read_only(self):
        """Test DELETE with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url,
            self.superuser,
            204,
            method='DELETE',
            cleanup_method=self._cleanup_delete,
        )
        self.assert_response_api(
            self.url, self.auth_non_superusers, 403, method='DELETE'
        )
        self.assert_response_api(self.url, self.anonymous, 401, method='DELETE')



# Generated by Django 4.2.22 on 2025-06-19 10:42

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0035_populate_viewer_role'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='public_access',
            field=models.ForeignKey(
                blank=True,
                help_text='Allow public access of specific read-only access '
                'level for project. Includes access for unauthenticated users '
                'if allowed on site.',
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='public_access',
                to='projectroles.role',
            ),
        ),
    ]



"""
Syncmodifyapi management command for synchronizing existing projects using the
project modify API.
"""

import sys

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import Project, SODAR_CONSTANTS
from projectroles.views import ProjectModifyPluginViewMixin


logger = ManagementCommandLogger(__name__)
User = get_user_model()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']


class Command(ProjectModifyPluginViewMixin, BaseCommand):
    help = 'Synchronizes existing projects using the project modify API'

    def add_arguments(self, parser):
        parser.add_argument(
            '-p',
            '--project',
            metavar='UUID',
            type=str,
            help='Limit sync to a project',
        )

    def handle(self, *args, **options):
        """Run management command"""
        project_uuid = options.get('project')
        if project_uuid:
            logger.info(
                f'Limiting sync to project with UUID "{project_uuid}"..'
            )
            project = Project.objects.filter(sodar_uuid=project_uuid).first()
            if not project:
                logger.error('Project not found')
                sys.exit(1)
            project_list = [project]
        else:
            logger.info('Synchronizing all projects..')
            project_list = Project.objects.all().order_by('full_title')
            logger.debug(
                f'Found {project_list.count()} projects and categories'
            )
        sync_count = 0
        err_count = 0
        for p in project_list:
            p_title = p.get_log_title(full_title=True)
            logger.debug(f'Synchronizing project {p_title}')
            try:
                self.call_project_modify_api('perform_project_sync', None, [p])
                sync_count += 1
            except Exception as ex:
                logger.error(f'Exception for project {p_title}: {ex}')
                err_count += 1
        logger.info(
            'Project data synchronized ({} OK, {} error{})'.format(
                sync_count, err_count, 's' if err_count != 1 else ''
            )
        )



"""Tests for AppLinkAPI in the projectroles app"""

from django.contrib.auth.models import AnonymousUser
from django.test import override_settings
from django.urls import reverse

from projectroles.app_links import AppLinkAPI
from projectroles.app_settings import AppSettingAPI
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_models import ProjectMixin, RoleAssignmentMixin
from projectroles.tests.test_views import ViewTestBase


app_links = AppLinkAPI()
app_settings = AppSettingAPI()


# SODAR constants
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']

# Local constants
APP_NAME = 'projectroles'
APP_NAME_FF = 'filesfolders'


class TestAppLinkAPI(ProjectMixin, RoleAssignmentMixin, ViewTestBase):
    """Tests for API"""

    def setUp(self):
        super().setUp()
        self.user_owner = self.make_user('user_owner')
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )

    def test_get_project_links(self):
        """Test get_project_links() with project"""
        kw = {'project': self.project.sodar_uuid}
        expected = [
            {
                'name': 'project-detail',
                'url': reverse('projectroles:detail', kwargs=kw),
                'label': 'Project Overview',
                'icon': 'mdi:cube',
                'active': False,
            },
            {
                'name': 'app-plugin-filesfolders',
                'url': reverse('filesfolders:list', kwargs=kw),
                'label': 'Files',
                'icon': 'mdi:file',
                'active': False,
            },
            {
                'name': 'app-plugin-timeline',
                'url': reverse('timeline:list_project', kwargs=kw),
                'label': 'Timeline',
                'icon': 'mdi:clock-time-eight',
                'active': False,
            },
            {
                'name': 'app-plugin-bgjobs',
                'url': reverse('bgjobs:list', kwargs=kw),
                'label': 'Background Jobs',
                'icon': 'mdi:server',
                'active': False,
            },
            {
                'name': 'app-plugin-example_project_app',
                'url': reverse('example_project_app:example', kwargs=kw),
                'label': 'Example Project App',
                'icon': 'mdi:rocket-launch',
                'active': False,
            },
            {
                'name': 'project-roles',
                'url': reverse('projectroles:roles', kwargs=kw),
                'label': 'Members',
                'icon': 'mdi:account-multiple',
                'active': False,
            },
            {
                'name': 'project-update',
                'url': reverse('projectroles:update', kwargs=kw),
                'label': 'Update Project',
                'icon': 'mdi:lead-pencil',
                'active': False,
            },
        ]
        self.assertEqual(
            app_links.get_project_links(self.user_owner, self.project), expected
        )

    def test_get_project_links_app_name(self):
        """Test get_project_links() with project and specific app plugin"""
        kw = {'project': self.project.sodar_uuid}
        expected = [
            {
                'name': 'project-detail',
                'url': reverse('projectroles:detail', kwargs=kw),
                'label': 'Project Overview',
                'icon': 'mdi:cube',
                'active': False,
            },
            {
                'name': 'app-plugin-filesfolders',
                'url': reverse('filesfolders:list', kwargs=kw),
                'label': 'Files',
                'icon': 'mdi:file',
                'active': True,  # This should be active
            },
            {
                'name': 'app-plugin-timeline',
                'url': reverse('timeline:list_project', kwargs=kw),
                'label': 'Timeline',
                'icon': 'mdi:clock-time-eight',
                'active': False,
            },
            {
                'name': 'app-plugin-bgjobs',
                'url': reverse('bgjobs:list', kwargs=kw),
                'label': 'Background Jobs',
                'icon': 'mdi:server',
                'active': False,
            },
            {
                'name': 'app-plugin-example_project_app',
                'url': reverse('example_project_app:example', kwargs=kw),
                'label': 'Example Project App',
                'icon': 'mdi:rocket-launch',
                'active': False,
            },
            {
                'name': 'project-roles',
                'url': reverse('projectroles:roles', kwargs=kw),
                'label': 'Members',
                'icon': 'mdi:account-multiple',
                'active': False,
            },
            {
                'name': 'project-update',
                'url': reverse('projectroles:update', kwargs=kw),
                'label': 'Update Project',
                'icon': 'mdi:lead-pencil',
                'active': False,
            },
        ]
        self.assertEqual(
            app_links.get_project_links(
                self.user_owner, self.project, app_name=APP_NAME_FF
            ),
            expected,
        )

    def test_get_project_links_category(self):
        """Test get_project_links() with category"""
        kw = {'project': self.category.sodar_uuid}
        expected = [
            {
                'name': 'project-detail',
                'url': reverse('projectroles:detail', kwargs=kw),
                'label': 'Category Overview',
                'icon': 'mdi:rhombus-split',
                'active': False,
            },
            {
                'name': 'app-plugin-timeline',
                'url': reverse('timeline:list_project', kwargs=kw),
                'label': 'Timeline',
                'icon': 'mdi:clock-time-eight',
                'active': False,
            },
            {
                'name': 'project-roles',
                'url': reverse('projectroles:roles', kwargs=kw),
                'label': 'Members',
                'icon': 'mdi:account-multiple',
                'active': False,
            },
            {
                'name': 'project-update',
                'url': reverse('projectroles:update', kwargs=kw),
                'label': 'Update Category',
                'icon': 'mdi:lead-pencil',
                'active': False,
            },
            {
                'name': 'project-create',
                'url': reverse('projectroles:create', kwargs=kw),
                'label': 'Create Project or Category',
                'icon': 'mdi:plus-thick',
                'active': False,
            },
        ]
        self.assertEqual(
            app_links.get_project_links(self.user_owner, self.category),
            expected,
        )

    def test_get_project_links_home_superuser(self):
        """Test get_project_links() with home URL as superuser"""
        expected = [
            {
                'name': 'home-project-create',
                'url': reverse('projectroles:create'),
                'label': 'Create Category',
                'icon': 'mdi:plus-thick',
                'active': False,
            },
        ]
        self.assertEqual(
            app_links.get_project_links(self.user, url_name='home'),
            expected,
        )

    def test_get_project_links_home_regular_user(self):
        """Test get_project_links() with home URL as regular user"""
        self.assertEqual(
            app_links.get_project_links(self.user_owner, url_name='home'), []
        )

    def test_get_project_links_url_name_projectroles(self):
        """Test get_project_links() with projectroles URL name"""
        links = app_links.get_project_links(
            self.user_owner,
            self.project,
            app_name=APP_NAME,
            url_name='roles',
        )
        self.assertEqual(len(links), 7)
        for i in range(0, 6):
            if i == 5:
                self.assertEqual(links[i]['name'], 'project-roles')
                self.assertEqual(links[i]['active'], True)
            else:
                self.assertEqual(links[i]['active'], False)

    def test_get_project_links_url_name_app_plugin(self):
        """Test get_project_links() with app plugin URL name"""
        links = app_links.get_project_links(
            self.user_owner,
            self.project,
            app_name=APP_NAME_FF,
            url_name='file_create',
        )
        self.assertEqual(len(links), 7)
        for i in range(0, 6):
            if i == 1:
                self.assertEqual(links[i]['name'], 'app-plugin-filesfolders')
                self.assertEqual(links[i]['active'], True)
            else:
                self.assertEqual(links[i]['active'], False)

    def test_get_project_links_contributor(self):
        """Test get_project_links() with project as contributor"""
        user_new = self.make_user('user_new')
        self.make_assignment(self.project, user_new, self.role_contributor)
        links = app_links.get_project_links(user_new, self.project)
        self.assertEqual(len(links), 6)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-update', link_names)

    def test_get_project_links_guest(self):
        """Test get_project_links() with project as guest"""
        user_new = self.make_user('user_new')
        self.make_assignment(self.project, user_new, self.role_guest)
        links = app_links.get_project_links(user_new, self.project)
        self.assertEqual(len(links), 6)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-update', link_names)

    def test_get_project_links_category_contributor(self):
        """Test get_project_links() with category as contributor"""
        user_new = self.make_user('user_new')
        self.make_assignment(self.category, user_new, self.role_contributor)
        links = app_links.get_project_links(user_new, self.category)
        self.assertEqual(len(links), 4)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-update', link_names)  # NOTE: Can create

    def test_get_project_links_category_guest(self):
        """Test get_project_links() with category as guest"""
        user_new = self.make_user('user_new')
        self.make_assignment(self.category, user_new, self.role_guest)
        links = app_links.get_project_links(user_new, self.category)
        self.assertEqual(len(links), 3)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-update', link_names)
        self.assertNotIn('project-create', link_names)

    @override_settings(PROJECTROLES_HIDE_PROJECT_APPS=[APP_NAME_FF])
    def test_get_project_links_hidden_app(self):
        """Test get_project_links() with hidden app"""
        links = app_links.get_project_links(self.user, self.project)
        self.assertEqual(len(links), 6)
        link_names = [link['name'] for link in links]
        self.assertNotIn('app-plugin-filesfolders', link_names)

    @override_settings(
        PROJECTROLES_SITE_MODE=SITE_MODE_TARGET,
        PROJECTROLES_TARGET_CREATE=False,
    )
    def test_get_project_links_category_target_disallow(self):
        """Test get_project_links() as target site with creation disallowed"""
        links = app_links.get_project_links(self.user_owner, self.category)
        self.assertEqual(len(links), 4)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-create', link_names)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_get_project_links_category_target_allow(self):
        """Test get_project_links() as target site with creation sallowed"""
        links = app_links.get_project_links(self.user_owner, self.category)
        self.assertEqual(len(links), 5)

    @override_settings(
        PROJECTROLES_SITE_MODE=SITE_MODE_TARGET,
        PROJECTROLES_TARGET_CREATE=False,
    )
    def test_get_project_links_category_target_disallow_superuser(self):
        """Test get_project_links() as superuser on target site with creation disallowed"""
        links = app_links.get_project_links(self.user, self.category)
        self.assertEqual(len(links), 4)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-create', link_names)

    def test_get_project_links_read_only(self):
        """Test get_project_links() with project and site read-only mode"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        links = app_links.get_project_links(self.user_owner, self.project)
        self.assertEqual(len(links), 6)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-update', link_names)

    def test_get_project_links_read_only_superuser(self):
        """Test get_project_links() with project and site read-only mode as superuser"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        links = app_links.get_project_links(self.user, self.project)
        self.assertEqual(len(links), 7)

    def test_get_project_links_category_read_only(self):
        """Test get_project_links() with category and site read-only mode"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        links = app_links.get_project_links(self.user_owner, self.category)
        self.assertEqual(len(links), 3)
        link_names = [link['name'] for link in links]
        self.assertNotIn('project-update', link_names)
        self.assertNotIn('project-create', link_names)

    def test_get_project_links_category_read_only_superuser(self):
        """Test get_project_links() with category and site read-only mode as superuser"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        links = app_links.get_project_links(self.user, self.category)
        self.assertEqual(len(links), 5)

    def test_get_user_links(self):
        """Test get_user_links() as regular user"""
        expected = [
            {
                'name': 'appalerts',
                'url': reverse('appalerts:list'),
                'label': 'App Alerts',
                'icon': 'mdi:alert-octagram',
                'active': False,
            },
            {
                'name': 'example_site_app',
                'url': reverse('example_site_app:example'),
                'label': 'Example Site App',
                'icon': 'mdi:rocket-launch-outline',
                'active': False,
            },
            {
                'name': 'timeline_site',
                'url': reverse('timeline:list_site'),
                'label': 'Site-Wide Events',
                'icon': 'mdi:clock-time-eight-outline',
                'active': False,
            },
            {
                'name': 'tokens',
                'url': reverse('tokens:list'),
                'label': 'API Tokens',
                'icon': 'mdi:key-chain-variant',
                'active': False,
            },
            {
                'name': 'userprofile',
                'url': reverse('userprofile:detail'),
                'label': 'User Profile',
                'icon': 'mdi:account-details',
                'active': False,
            },
            {
                'name': 'sign-out',
                'url': reverse('logout'),
                'label': 'Logout',
                'icon': 'mdi:logout-variant',
                'active': False,
            },
        ]
        self.assertEqual(app_links.get_user_links(self.user_owner), expected)

    def test_get_user_links_superuser(self):
        """Test get_user_links() as superuser"""
        expected = [
            {
                'name': 'adminalerts',
                'url': reverse('adminalerts:list'),
                'label': 'Admin Alerts',
                'icon': 'mdi:alert',
                'active': False,
            },
            {
                'name': 'appalerts',
                'url': reverse('appalerts:list'),
                'label': 'App Alerts',
                'icon': 'mdi:alert-octagram',
                'active': False,
            },
            {
                'name': 'bgjobs_site',
                'url': reverse('bgjobs:site_list'),
                'label': 'Site Background Jobs',
                'icon': 'mdi:server',
                'active': False,
            },
            {
                'name': 'example_site_app',
                'url': reverse('example_site_app:example'),
                'label': 'Example Site App',
                'icon': 'mdi:rocket-launch-outline',
                'active': False,
            },
            {
                'name': 'remotesites',
                'url': reverse('projectroles:remote_sites'),
                'label': 'Remote Site Access',
                'icon': 'mdi:cloud-sync',
                'active': False,
            },
            {
                'name': 'siteappsettings',
                'url': '/project/site-app-settings',
                'label': 'Site App Settings',
                'icon': 'mdi:cog-outline',
                'active': False,
            },
            {
                'name': 'siteinfo',
                'url': reverse('siteinfo:info'),
                'label': 'Site Info',
                'icon': 'mdi:bar-chart',
                'active': False,
            },
            {
                'name': 'timeline_site',
                'url': reverse('timeline:list_site'),
                'label': 'Site-Wide Events',
                'icon': 'mdi:clock-time-eight-outline',
                'active': False,
            },
            {
                'name': 'timeline_site_admin',
                'url': reverse('timeline:list_admin'),
                'label': 'All Timeline Events',
                'icon': 'mdi:web-clock',
                'active': False,
            },
            {
                'name': 'tokens',
                'url': reverse('tokens:list'),
                'label': 'API Tokens',
                'icon': 'mdi:key-chain-variant',
                'active': False,
            },
            {
                'name': 'userprofile',
                'url': reverse('userprofile:detail'),
                'label': 'User Profile',
                'icon': 'mdi:account-details',
                'active': False,
            },
            {
                'name': 'admin',
                'url': None,  # No URL for Django admin, opens warning modal
                'label': 'Django Admin',
                'icon': 'mdi:cogs',
                'active': False,
            },
            {
                'name': 'sign-out',
                'url': reverse('logout'),
                'label': 'Logout',
                'icon': 'mdi:logout-variant',
                'active': False,
            },
        ]
        self.assertEqual(app_links.get_user_links(self.user), expected)

    def test_get_user_links_app_name(self):
        """Test get_user_links() with app plugin name"""
        links = app_links.get_user_links(self.user_owner, app_name='tokens')
        self.assertEqual(len(links), 6)
        for i in range(0, 5):
            if i == 3:
                self.assertEqual(links[i]['name'], 'tokens')
                self.assertEqual(links[i]['active'], True)
            else:
                self.assertEqual(links[i]['active'], False)

    def test_get_user_links_url_name(self):
        """Test get_user_links() with URL name"""
        links = app_links.get_user_links(
            self.user_owner, app_name='tokens', url_name='create'
        )
        self.assertEqual(len(links), 6)
        for i in range(0, 5):
            if i == 3:
                self.assertEqual(links[i]['name'], 'tokens')
                self.assertEqual(links[i]['active'], True)
            else:
                self.assertEqual(links[i]['active'], False)

    def test_get_user_links_url_name_remote(self):
        """Test get_user_links() with remote sites URL name"""
        links = app_links.get_user_links(
            self.user, app_name=APP_NAME, url_name='remote_site_create'
        )
        self.assertEqual(len(links), 13)
        for i in range(0, 13):
            if i == 4:
                self.assertEqual(links[i]['name'], 'remotesites')
                self.assertEqual(links[i]['active'], True)
            else:
                self.assertEqual(links[i]['active'], False)

    def test_get_user_links_url_name_site_app_settings(self):
        """Test get_user_links() with site app settings URL name"""
        links = app_links.get_user_links(
            self.user, app_name=APP_NAME, url_name='site_app_settings'
        )
        self.assertEqual(len(links), 13)
        for i in range(0, 13):
            if i == 5:
                self.assertEqual(links[i]['name'], 'siteappsettings')
                self.assertEqual(links[i]['active'], True)
            else:
                self.assertEqual(links[i]['active'], False)

    def test_get_user_links_anon(self):
        """Test get_user_links() as anonymous user"""
        links = app_links.get_user_links(AnonymousUser())
        self.assertEqual(len(links), 1)
        self.assertEqual(links[0]['name'], 'sign-in')

    @override_settings(PROJECTROLES_KIOSK_MODE=True)
    def test_get_user_links_anon_kiosk_mode(self):
        """Test get_user_links() as anonymous user and kiosk mode"""
        links = app_links.get_user_links(AnonymousUser())
        self.assertEqual(len(links), 0)

    def test_get_user_links_read_only(self):
        """Test get_user_links() with site read-only mode as regular user"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        # All should be returned
        self.assertEqual(len(app_links.get_user_links(self.user_owner)), 6)

    def test_get_user_links_read_only_superuser(self):
        """Test get_user_links() with site read-only mode as superuser"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        self.assertEqual(len(app_links.get_user_links(self.user)), 13)



"""Forms for the filesfolders app"""

from typing import Optional
from zipfile import ZipFile

from django import forms
from django.conf import settings
from django.template.defaultfilters import filesizeformat

from db_file_storage.form_widgets import DBClearableFileInput

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.forms import SODARModelForm
from projectroles.models import Project, SODARUser
from projectroles.utils import build_secret

from filesfolders.models import File, Folder, HyperLink


app_settings = AppSettingAPI()


# Settings
MAX_UPLOAD_SIZE = settings.FILESFOLDERS_MAX_UPLOAD_SIZE
MAX_ARCHIVE_SIZE = settings.FILESFOLDERS_MAX_ARCHIVE_SIZE

# Local constants
APP_NAME = 'filesfolders'


class FilesfoldersItemForm(SODARModelForm):
    """Base form for Filesfolders item creation/updating"""

    def __init__(
        self, current_user=None, folder=None, project=None, *args, **kwargs
    ):
        """Override for form initialization"""
        super().__init__(*args, **kwargs)
        self.current_user = None
        self.project = None
        self.folder = None
        # Get current user for checking permissions for form items
        if current_user:
            self.current_user = current_user
        if folder:
            self.folder = Folder.objects.get(sodar_uuid=folder)
            self.project = self.folder.project
        elif project:
            self.project = Project.objects.get(sodar_uuid=project)
        # Modify ModelChoiceFields to use sodar_uuid
        self.fields['folder'].to_field_name = 'sodar_uuid'


class FolderForm(FilesfoldersItemForm):
    """Form for Folder creation/updating"""

    class Meta:
        model = Folder
        fields = ['name', 'folder', 'flag', 'description']

    def __init__(
        self,
        current_user: Optional[SODARUser] = None,
        folder: Optional[Folder] = None,
        project: Optional[Project] = None,
        *args,
        **kwargs,
    ):
        """Override for form initialization"""
        super().__init__(
            current_user=current_user,
            project=project,
            folder=folder,
            *args,
            **kwargs,
        )

        # Creation
        if not self.instance.pk:
            # Don't allow changing folder if we are creating a new object
            self.initial['folder'] = (
                self.folder.sodar_uuid if self.folder else None
            )
            self.fields['folder'].widget = forms.HiddenInput()

        # Updating
        else:
            # Allow moving folder inside other folders in project
            folder_choices = [(None, 'root')]
            folders = Folder.objects.filter(
                project=self.instance.project.pk
            ).exclude(pk=self.instance.pk)

            # Exclude everything under current folder
            folders = [f for f in folders if not f.has_in_path(self.instance)]
            for f in folders:
                folder_choices.append((f.sodar_uuid, f.get_path()))

            self.fields['folder'].choices = folder_choices
            self.initial['folder'] = (
                self.instance.folder.sodar_uuid
                if self.instance.folder
                else None
            )

    def clean(self):
        # Creation
        if not self.instance.pk:
            try:
                Folder.objects.get(
                    project=self.project,
                    folder=self.folder,
                    name=self.cleaned_data['name'],
                )
                self.add_error('name', 'Folder already exists')
            except Folder.DoesNotExist:
                pass

        # Updating
        else:
            # Ensure a folder with the same name does not exist in the location
            old_folder = None
            try:
                old_folder = Folder.objects.get(pk=self.instance.pk)
            except Folder.DoesNotExist:
                pass
            if old_folder and (
                old_folder.name != self.cleaned_data['name']
                or old_folder.folder != self.cleaned_data['folder']
            ):
                try:
                    Folder.objects.get(
                        project=self.instance.project,
                        folder=self.cleaned_data['folder'],
                        name=self.cleaned_data['name'],
                    )
                    self.add_error('name', 'Folder already exists')
                except Folder.DoesNotExist:
                    pass
        return self.cleaned_data

    def save(self, *args, **kwargs):
        """Override of form saving function"""
        obj = super().save(commit=False)
        # Updating
        if self.instance.pk:
            obj.owner = self.instance.owner
            obj.project = self.instance.project
            obj.folder = self.instance.folder
        # Creation
        else:
            obj.owner = self.current_user
            obj.project = self.project
            if self.folder:
                obj.folder = self.folder
        obj.save()
        return obj


class FileForm(FilesfoldersItemForm):
    """Form for File creation/updating"""

    unpack_archive = forms.BooleanField(
        required=False, label='Extract files from archive'
    )

    class Meta:
        model = File
        fields = [
            'file',
            'unpack_archive',
            'folder',
            'description',
            'flag',
            'public_url',
        ]
        widgets = {'file': DBClearableFileInput}
        help_texts = {
            'file': f'Uploaded file (maximum size: '
            f'{filesizeformat(MAX_UPLOAD_SIZE)})'
        }

    @staticmethod
    def _get_file_size(file) -> int:
        try:
            return file.size
        except NotImplementedError:
            return file.file.size

    def _check_size(self, file_size: int, limit: int) -> bool:
        if file_size > limit:
            self.add_error(
                'file',
                f'File too large, maximum size is {limit} bytes '
                f'(file size is {file_size} bytes)',
            )
            return False
        return True

    def __init__(
        self,
        current_user: Optional[SODARUser] = None,
        folder: Optional[Folder] = None,
        project: Optional[Project] = None,
        *args,
        **kwargs,
    ):
        """Override for form initialization"""
        super().__init__(
            current_user=current_user,
            folder=folder,
            project=project,
            *args,
            **kwargs,
        )
        if self.instance.pk:
            self.project = self.instance.project
        self.fields['public_url'].label = 'Public link'

        # Disable public URL creation if setting is false
        if not app_settings.get(
            APP_NAME, 'allow_public_links', project=self.project
        ):
            self.fields['public_url'].disabled = True

        # Creation
        if not self.instance.pk:
            # Don't allow changing folder if we are creating a new object
            self.initial['folder'] = (
                self.folder.sodar_uuid if self.folder else None
            )
            self.fields['folder'].widget = forms.HiddenInput()
            self.fields['file'].required = True

        # Updating
        else:
            # Allow moving file inside other folders in project
            folder_choices = [(None, 'root')]
            for f in Folder.objects.filter(project=self.instance.project.pk):
                folder_choices.append((f.sodar_uuid, f.get_path()))
            self.fields['folder'].choices = folder_choices
            self.initial['folder'] = (
                self.instance.folder.sodar_uuid
                if self.instance.folder
                else None
            )

    def clean(self):
        project = self.instance.project if self.instance.pk else self.project
        folder = self.cleaned_data.get('folder')
        file = self.cleaned_data.get('file')
        if not file:
            return self.cleaned_data

        unpack_archive = self.cleaned_data.get('unpack_archive')
        new_filename = file.name.split('/')[-1]
        size = self._get_file_size(file)

        # Normal file handling
        if file and (
            not hasattr(file, 'content_type')
            or file.content_type
            not in ['application/zip', 'application/x-zip-compressed']
        ):
            # Ensure max file size is not exceeded
            if not self._check_size(size, MAX_UPLOAD_SIZE):
                return self.cleaned_data
            # Attempting to unpack a non-zip file
            if unpack_archive:
                self.add_error(
                    'unpack_archive',
                    'Attempting to extract from a file that is not a '
                    'Zip archive',
                )
                return self.cleaned_data

        # Zip archive handling
        elif unpack_archive:
            # Ensure max archive size is not exceeded
            if not self._check_size(size, MAX_ARCHIVE_SIZE):
                return self.cleaned_data
            try:
                zip_file = ZipFile(file)
            except Exception as ex:
                self.add_error('file', f'Unable to open zip file: {ex}')
                return self.cleaned_data
            archive_files = [f for f in zip_file.infolist() if not f.is_dir()]
            if len(archive_files) == 0:
                self.add_error(
                    'file', 'Found nothing to extract from zip archive'
                )
                return self.cleaned_data

            for f in archive_files:
                # Ensure file size
                if not self._check_size(f.file_size, MAX_UPLOAD_SIZE):
                    return self.cleaned_data
                # Check if any of the files exist
                path_split = f.filename.split('/')
                check_folder = folder
                for p in path_split[:-1]:
                    # Advance in path
                    check_folder = Folder.objects.filter(
                        name=p, folder=check_folder
                    ).first()
                # Once reached the correct path, check if file exists
                if File.objects.filter(
                    name=path_split[-1], folder=check_folder
                ).first():
                    self.add_error('file', f'File already exists: {f.filename}')
                    return self.cleaned_data

        # Creation
        if (
            not self.instance.pk
            and not unpack_archive
            and File.objects.filter(
                project=project, folder=self.folder, name=file.name
            ).first()
        ):
            self.add_error('file', 'File already exists')

        # Updating
        elif self.instance.pk:
            # Ensure file with the same name does not exist in the same
            # folder (unless we update file with the same folder and name)
            old_file = File.objects.filter(
                project=self.instance.project,
                folder=self.instance.folder,
                name=self.instance.name,
            ).first()
            if (
                old_file
                and self.instance.name != str(file)
                and File.objects.filter(
                    project=self.instance.project, folder=folder, name=file
                ).first()
            ):
                self.add_error('file', 'File already exists')
            # Moving:
            # If moving, ensure an identical file doesn't exist in the
            # target folder
            if (
                self.instance.folder != folder
                and File.objects.filter(
                    project=self.instance.project,
                    folder=folder,
                    name__in=[new_filename, self.instance.name],
                ).count()
                > 0
            ):
                self.add_error(
                    'folder',
                    'File with identical name already exists in folder',
                )
        return self.cleaned_data

    def save(self, *args, **kwargs):
        """Override of form saving function"""
        obj = super().save(commit=False)
        # Creation
        if not self.instance.pk:
            obj.name = obj.file.name
            obj.owner = self.current_user
            obj.project = self.project
            if self.folder:
                obj.folder = self.folder
            obj.secret = build_secret()  # Secret string created here
        # Updating
        else:
            old_file = File.objects.get(pk=self.instance.pk)
            if old_file.file != self.instance.file:
                obj.file = self.instance.file
                obj.name = obj.file.name.split('/')[-1]

            obj.owner = self.instance.owner
            obj.project = self.instance.project
            obj.secret = self.instance.secret
            if self.instance.folder:
                obj.folder = self.instance.folder

            if app_settings.get(
                APP_NAME, 'allow_public_links', project=self.instance.project
            ):
                obj.public_url = self.instance.public_url
            else:
                obj.public_url = False
        obj.save()
        return obj


class HyperLinkForm(FilesfoldersItemForm):
    """Form for HyperLink creation/updating"""

    class Meta:
        model = HyperLink
        fields = ['name', 'url', 'folder', 'flag', 'description']

    def __init__(
        self,
        current_user: Optional[SODARUser] = None,
        folder: Optional[Folder] = None,
        project: Optional[Project] = None,
        *args,
        **kwargs,
    ):
        """Override for form initialization"""
        super().__init__(
            current_user=current_user,
            project=project,
            folder=folder,
            *args,
            **kwargs,
        )
        # Creation
        if not self.instance.pk:
            # Don't allow changing folder if we are creating a new object
            self.initial['folder'] = (
                self.folder.sodar_uuid if self.folder else None
            )
            self.fields['folder'].widget = forms.HiddenInput()
        # Updating
        else:
            # Allow moving file inside other folders in project
            folder_choices = [(None, 'root')]
            for f in Folder.objects.filter(project=self.instance.project.pk):
                folder_choices.append((f.sodar_uuid, f.get_path()))
            self.fields['folder'].choices = folder_choices
            self.initial['folder'] = (
                self.instance.folder.sodar_uuid
                if self.instance.folder
                else None
            )

    def clean(self):
        # Creation
        if not self.instance.pk:
            try:
                HyperLink.objects.get(
                    project=self.project,
                    folder=self.folder,
                    name=self.cleaned_data['name'],
                )
                self.add_error('name', 'Link already exists')
            except HyperLink.DoesNotExist:
                pass
        # Updating
        else:
            # Ensure a link with the same name does not exist in the location
            old_link = None
            try:
                old_link = HyperLink.objects.get(pk=self.instance.pk)
            except HyperLink.DoesNotExist:
                pass
            if old_link and (
                old_link.name != self.cleaned_data['name']
                or old_link.folder != self.cleaned_data['folder']
            ):
                try:
                    HyperLink.objects.get(
                        project=self.instance.project,
                        folder=self.cleaned_data['folder'],
                        name=self.cleaned_data['name'],
                    )
                    self.add_error('name', 'Link already exists')
                except HyperLink.DoesNotExist:
                    pass
        return self.cleaned_data

    def save(self, *args, **kwargs):
        """Override of form saving function"""
        obj = super().save(commit=False)
        # Updating
        if self.instance.pk:
            obj.owner = self.instance.owner
            obj.project = self.instance.project
            obj.folder = self.instance.folder
        # Creation
        else:
            obj.owner = self.current_user
            obj.project = self.project
            if self.folder:
                obj.folder = self.folder
        obj.save()
        return obj
